package state51::Odoo;
use Moose;

use Carp 'croak';
use Frontier::Client;
use MojoX::JSONRPC2::HTTP;
use Moose::Util::TypeConstraints;
use WWW::Mechanize;
use state51::Logger;
use state51::Odoo::Types;
use state51::Odoo::StockManagement::Constants qw/
    ODOO_WAREHOUSE_INVENTORY_ADJUSTMENT_TABLE
    ODOO_WAREHOUSE_PRODUCT_TABLE
    ODOO_WAREHOUSE_STOCK_MOVE_STOCK_QUANT_IDS_FIELD
    ODOO_WAREHOUSE_STOCK_MOVE_TABLE
/;

our $VERSION = '2024091700';

=head1 NAME

state51::Odoo

=head1 DESCRIPTION

state51::Odoo is a library for working with Odoo API. Odoo provides two
interfaces, XML-RPC and JSON-RPC that are both implemented by our module

=head1 SYNOPSIS

    use state51::Odoo;

    # For the XML-RPC interface
    my $odoo = state51::Odoo->new();
    my $uid = $odoo->odoo_uid;

    my $got = $odoo->execute(
            'account.move',
            'search',
            [
                [ 'ref', '>=',   $self->invoice_prefix . $invoice_number ],
                [ 'ref', 'like', $prefix . "%" ]
            ]
    );

    # For the JSON-RPC interface
    my $odoo = state51::Odoo->new(interface_type => 'json');

=head2 Debuging

It's highly recommended to activate the debug flags to display generated XML /
JSON and possibly other client's API interactions feedback.

=over 2

=item XML-RPC

  # Should be hardcoded in state51::Odoo, but would be a nice ENV addition
  Frontier::Client->new('url' => $server_url, debug => 1);

=item JSON-RPC

  C<MOJO_CLIENT_DEBUG=1>

=back

=cut

with 'state51::Odoo::Role::JSONRPC';

use constant ODOO_TIMEOUT_SEC => $ENV{ODOO_TIMEOUT_SEC} // 180;

has user => (
    is      => 'ro',
    isa     => 'Str',
    default => sub {
        return $ENV{ODOO_USERNAME} || die('No ODOO_USERNAME environment');
    },
);

has password => (
    is      => 'ro',
    isa     => 'Str',
    default => sub {
        return $ENV{ODOO_PASSWORD} || die('No ODOO_PASSWORD environment');
    },
);

has db_name => (
    is      => 'ro',
    isa     => 'Str',
    default => sub {
        return $ENV{ODOO_DB} || die('No ODOO_DB environment');
    },
);

has base_url => (
    is      => 'ro',
    isa     => 'Str',
    default => sub {
        my $base = $ENV{ODOO_URL} || die('No ODOO_URL environment');
        $base =~ s/\/$//;
        return $base;
    },
);

has logger => (
    is  => 'ro',
    isa => 'state51::Logger',
    default => sub {
        state51::Logger->new();
    },
);

=head2 interface_type

Odoo is accessible through XML-RPC/JSON-RPC interfaces. Defaults to XML-RPC, but
can also be JSON-RPC.

    # Defaults to 'xml' state51::Odoo->interface_type
    state51::Odoo->new();

    # Force JSON-RPC usage
    state51::Odoo->new(interface_type => 'json');

Any other value is forbidden (enforced by the C<XMLOrJSONInterface> C<enum>).

=cut

has interface_type => (
    is      => 'ro',
    isa     => 'XMLOrJSONInterface',
    default => 'xml',
);

has common_service => (
    is      => 'ro',
    isa     => 'XMLRPCClass | JSONRPCClass',
    lazy    => 1,
    builder => 'build_common_service',
);

=head2 build_common_service

Lazy builder to instantiate RPC client instance for common operations. The
JSON-RPC client made with Frontier::RPC behave very differently from
MojoX::JSONRPC2::HTTP, so note that the the used path should look like the
following.

=over 2

=item XML-RPC

    https://odoo-dev-8069.int.openimp.com/jsonrpc/common

=item JSON-RPC

    https://odoo-dev-8069.int.openimp.com/jsonrpc

=back

=cut

# TODO We missing unit tests
sub build_common_service {
    my ($self) = @_;

    my $interface_path = '/' . $self->interface_type . 'rpc';
    $interface_path .= '/common' unless $self->interface_type eq 'json';

    my $common_service = $self->build_service($interface_path);

    return $common_service;
}

has object_service => (
    is      => 'ro',
    isa     => 'XMLRPCClass | JSONRPCClass',
    lazy    => 1,
    builder => 'build_object_service',
);

=head2 build_object_service

Lazy builder to instantiate xml_rpc client instance for object operations.

=cut

# TODO Missing unit tests
sub build_object_service {
    my ($self) = @_;
    my $interface_path = '/' . $self->interface_type . 'rpc';
    $interface_path .= '/object' unless $self->interface_type eq 'json';
    my $object_service = $self->build_service($interface_path);
    return $object_service;
}

sub build_service {
    my ($self, $path) = @_;

    my $server_url = $self->base_url . $path;

    my $cl;

    if ($self->interface_type eq 'xml') {
        $cl = Frontier::Client->new('url' => $server_url);
        $cl->{ua}{timeout} = ODOO_TIMEOUT_SEC;
    }

    elsif ($self->interface_type eq 'json') {
        $cl = MojoX::JSONRPC2::HTTP->new({ url => $server_url, });

        $cl->ua->request_timeout(ODOO_TIMEOUT_SEC);
    }

    else {
        croak 'Abnormal situation with the Odoo API interface type';
    }

    return $cl;
}

has odoo_uid => (
    is            => 'ro',
    isa           => 'Int',
    lazy          => 1,
    builder       => 'login',
    documentation => 'logged in user_id',
    clearer       => 'odoo_sign_out',
);

=head2 login

Logging in to ODOO system. Would return an uid on success. On failure would
either return a

=over 2

=item XML-RPC

A false Frontier::RCP2::Boolean

=item JSON-RPC

A fatal error with detailed diagnostics. See state51::Odoo::Role::JSONRPC
C<_check_errors> doc for details.

=back

=cut

sub login {
    my ($self) = @_;

    if ($self->interface_type eq 'xml') {
        my $res = $self->common_service->call(
            'login',
            $self->db_name,
            $self->user,
            $self->password
        );

        if (ref $res eq 'Frontier::RPC2::Boolean') {
            $self->logger->message(
                level => 'warn',
                message => [
                    '[Odoo::login] '
                    . 'Something is wrong with the Odoo login, check the parameters.'
                ]
            ) if not $res->value;
        }
        return $res;
    }

    if ($self->interface_type eq 'json') {

        my %params = (
            args    => [ $self->db_name, $self->user, $self->password, ],
            service => 'common',
            method  => 'login',
        );

        my ($fail, $res, $err) = $self->common_service->call_named('call', %params,);

        croak "Transport error. Something wrong happened before you even reached the API: $fail" if $fail;
        croak "Something's wrong with the authentication: $err"                                  if not $res;

        return $res;
    }
}

has report_service => (
    is      => 'ro',
    isa     => 'Frontier::Client',
    lazy    => 1,
    builder => 'build_report_service',
);

=head2 build_report_service

Lazy builder to instantiate xml_rpc client instance for report operations.

=cut

sub build_report_service {
    my ($self) = @_;

    my $server_url = $self->base_url . '/report';
    return Frontier::Client->new('url' => $server_url);
}

has www_mechanize => (
    is      => 'rw',
    isa     => 'WWW::Mechanize',
    lazy    => 1,
    builder => 'build_www_mechanize',
);

sub build_www_mechanize {
    my ($self) = @_;

    my $mech = WWW::Mechanize->new(
        onerror => undef,
        agent   => 'Linux Mozilla'
    );

    my $login_url = $self->base_url . "/web/login";
    $mech->get($login_url);

    if ($mech->res->is_success) {
        my $csrf_token = $mech->value('csrf_token');

        my $resp = $mech->submit_form(
            with_fields => {
                "login"      => $self->user,
                "password"   => $self->password,
                "csrf_token" => $csrf_token,
            }
        );

        unless ($resp->is_success) {
            die "Web login failed";
        }
    }
    else {
        die $mech->content;
    }

    return $mech;
}

sub download_invoice_report {
    my ($self, $report_filename, $odoo_invoice_id) = @_;

    my $report_path = '/report/pdf/account.report_invoice/' . $odoo_invoice_id; 
    my $report_url = $self->base_url . $report_path;

    my $response = $self->www_mechanize->get($report_url, ":content_file" => $report_filename);

    unless ($response->is_success) {
        $self->logger->message(
            level => 'warn',
            message => [
                '[Odoo::download_invoice_report] '
                . 'The invoice report file could not be downloaded. '
                . 'Code: ' . $response->code
                . 'Message: ' . $response->message
            ]
        );

        return;
    }

    return $response;
}

sub execute {
    my ($self, $model, $method, $position, $keywords, $args) = @_;

    unless ($model && $method) {
        $self->logger->message(
            level => 'warn',
            message => [
                '[Odoo::execute] Model or method is empty',
            ]
        );

        return;
    }

    my $result = eval {
        $self->object_service->call(
            'execute', $self->db_name, $self->odoo_uid, $self->password,
            $model, $method, $position,
            $keywords ? $keywords : (),
        );
    };

    if ($@) {
        $self->logger->message(
            level => 'fatal',
            message => ['[Odoo::execute] ' . $@],
        );

        return;
    }

    if ($method eq 'read' and $args->{replace_boolean}) {
        for my $item (@$result) {
            for my $k (keys %$item) {
                if (ref $item->{$k} eq 'Frontier::RPC2::Boolean') {
                    $item->{$k} = $item->{$k}->value || undef;
                }
            }
        }
    }

    return $result;
}

1;
