package state51::Odoo::Role::JSONRPC;

use v5.26;
use Moose::Role;
use Carp qw/carp croak/;
use List::Util qw(reduce);
use Time::Piece;

use state51::Odoo::StockManagement::Constants qw/
    ODOO_WAREHOUSE_INVENTORY_ADJUSTMENT_TABLE
    ODOO_WAREHOUSE_PRODUCT_TABLE
    ODOO_WAREHOUSE_STOCK_MOVE_STOCK_QUANT_IDS_FIELD
    ODOO_WAREHOUSE_STOCK_MOVE_TABLE
    /;

=head1 DESCRIPTION

Provide necessary features for the Odoo API JSON-RPC interface.

=head1 WARNING

Data "types" can matter with JSON-RPC: if you pass the string version of an
integer to a C<search_read> method, it will return nothing without warning just
because it was unable to find a C<"12345"> id (a JSON string), when it should
have been C<12345>. Makes sense in JSON context, you have been warned

=head1 TODO

  - The XML-RPC / JSON-RPC implementations should lay in dedicated classes / roles
  - Make ->execute() specific methods for each interface
  - Factorize the call execution this is too much boilerplate

=head1 METHODS

=head2 get_product()

Passing an optional boolean will return the stock_quant_ids
only, although leaving the parameter undef will return the whole Product object
(and yes, it can contain huge images).

=cut

sub get_product {

    my ($self, $product_id, $should_return_stock_quant_ids_only) = @_;

    # Fatal if it is not
    $self->_check_interface_type_is_json();

    my %params = (
        args => [
            $self->_aggregate_auth_params(),
            # Model
            ODOO_WAREHOUSE_PRODUCT_TABLE,
            'search_read',
            [
                # Filter, kind of WHERE id = $product_id
                [ 'id', '=', $product_id ],
            ],
            [
                qw/
                    id
                    property_stock_inventory
                    /
            ],

        ],
        service => 'object',
        method  => 'execute',
    );

    # If true, only return this field, not the rest of the Product object (it's huge, contains images)
    if ($should_return_stock_quant_ids_only) {
        # Kind of SELECT ODOO_WAREHOUSE_STOCK_MOVE_STOCK_QUANT_IDS_FIELD WHERE...
        push $params{args}->@*, [ODOO_WAREHOUSE_STOCK_MOVE_STOCK_QUANT_IDS_FIELD];
    }

    my ($fail, $res, $err) = $self->object_service->call_named('call', %params,);
    $self->_check_errors({ err => $err, res => $res, fail => $fail, });

    # NOTE That's an array, but there's only one item, that's fine
    my $product = $res->[0];

    # Also available:
    #
    #   barcode
    #   code
    #   stock_quant_ids
    #   description
    return {
        id          => $product->{id},
        location_id => $product->{property_stock_inventory}[0],
    };
}

sub get_product_stock_quant {

    my ($self, $product_id, $location_id) = @_;

    my %params = (
        args => [
            $self->_aggregate_auth_params(),
            # Model
            ODOO_WAREHOUSE_INVENTORY_ADJUSTMENT_TABLE,
            'search_read',
            [
                # Ids should be forced to integer representation, not strings, or the
                # search will return an empty array. See the WARNING POD section.
                [ 'product_id',  '=', $self->_normalise_sv_to_int($product_id ) ],
                [ 'location_id', '=', $self->_normalise_sv_to_int($location_id) ],
            ],
            [qw/id create_date/],
        ],
        service => 'object',
        method  => 'execute',
    );

    # XXX Need exceptions management here in case $res is empty so that the
    # script do not totally fail
    my ($fail, $res, $err) = $self->object_service->call_named('call', %params,);
    $self->_check_errors({ err => $err, res => $res, fail => $fail, action => 'none' });

    # TODO gl#12 if the stock.quant doesn't exist, we should create it
    if (not $res->@*) {
        carp sprintf('[warn] Missing stock.quant for %d location_id but we can deal with it', $location_id);
        return {};
    }

    # FIXME Possibly not needed as the product_id / location_id search should return only one line
    # If the array length is > 1, we got a problem, though
    my @stock_quant_with_most_recent_date_times = sort { $b->{create_date} cmp $a->{create_date} } $res->@*;

    return $stock_quant_with_most_recent_date_times[0];
}

sub do_reconciliation {
    my ($self, $adjustment) = @_;

    # TODO Create a tiny Moose object for that: there's a fair amount of data
    # validation and logic, I don't think we want to manage those "manually".
    # Could possibly be inspired state51::StockManagement::QuantityAdjustment
    #
    # The logic is defined in our specifications, see
    # https://docs.google.com/document/d/15Hej-js0Xj8hcdhnqPCtPOWTrOokIYyzX-oAlZZ_JvA/edit#heading=h.c0vqbs8doqrp
    #
    # Positive int, required
    my $odoo_product_id = $adjustment->{odoo_product_id};
    # Positive int, required
    my $stock_quant_id = $adjustment->{odoo_stock_quant_id};
    # Negative or positive int, required
    my $quantity_adjustment = $adjustment->{odoo_stock_quantity_adjustment};

    # XXX Locations I have no idea how to set them at the moment?
    # TODO location_id
    # TODO location_dest_id

    my $t        = localtime();
    my $odoo_now = $t->ymd . ' ' . $t->hms;

    # The methodology here is to UPDATE ('write' method) the stock.quant entry,
    # by setting the inventory_diff_quantity field. Then, we call the
    # action_apply_inventory action method and let Odoo process the data it's
    # own way. It will have the effect to insert the new history items
    # automatically (stock.move.line).
    my %params = (
        args => [
            $self->_aggregate_auth_params(),
            # Model
            ODOO_WAREHOUSE_INVENTORY_ADJUSTMENT_TABLE,
            'write',
            # That's the targeted product stock.quant id
            [ $stock_quant_id ],
            {
                inventory_diff_quantity => $quantity_adjustment,
                # TODO location_id      => $location_id,
                # TODO location_dest_id => $location_dest_id,
                # TODO owner_id         => '?', FIXME This is very much needed so that Odoo knows where it should 'take' the quantity from
            },
        ],
        service => 'object',
        method  => 'execute',
    );

    my ($fail, $res, $err) = $self->object_service->call_named('call', %params,);
    $self->_check_errors({ err => $err, res => $res, fail => $fail });

    say 'stock.quant record updated with status ' . $res;

    # Now we proceed to the method call that will trigger the data processing
    #
    # NOTE this is similar to clicking on the 'apply' button when setting the
    # new inventory quantity in the UI
    my %paramss = (
        args => [
            $self->_aggregate_auth_params(),
            # Model
            ODOO_WAREHOUSE_INVENTORY_ADJUSTMENT_TABLE,
            'action_apply_inventory',
            [ $stock_quant_id ],
        ],
        service => 'object',
        method  => 'execute',
    );

    my ($faill, $ress, $errr) = $self->object_service->call_named('call', %paramss,);
    $self->_check_errors({ err => $errr, res => $ress, fail => $faill, action => 'action_apply_inventory' });

    return 1;
}

sub _aggregate_auth_params {
    my $self   = shift;
    my @params = ($self->db_name, $self->odoo_uid, $self->password,);
    return @params;
}

sub _check_interface_type_is_json {
    my $self = shift;
    croak 'This method does not manage the XML-RPC interface. Please use JSON-RPC'
        if $self->interface_type ne 'json';
}

=head2 _check_errors()

In case of transport-level errors, when we fail to either send a RPC request or
receive a correct reply from RPC server the $fail will contain an error message,
while $res and $err will be undefined.

In case remote 'method' or RPC server itself will return error it will be
available in $err as an HASHREF with keys {code}, {message} and optionally {data},
while $fail and $res will be undefined. Otherwise value returned by remote
'method' will be in $res, while $fail and $err will be undefined.

See perldoc MojoX::JSONRPC2::HTTP for details

=cut

sub _check_errors {

    my ($self, $args) = @_;

    if (
        defined $args->{fail}
        and $args->{fail} eq 'expect {result} or {error}'
        # MojoX::JSONRPC2::HTTP assume a JSON::RPC server would always return
        # something: but the Odoo's implementation seem to return nothing when
        # calling this 'action'
        and $args->{action} eq 'action_apply_inventory'
        )
    {
        carp 'Nothing returned by the API when calling action_apply_inventory but that is fine';
        return;
    }

    if (not $args->{res}) {
        # XXX Could take advantage of Data::Printer here ?
        croak <<~"ERROR";
        Something is wrong with the Odoo API:
          message: $args->{err}->{message}
          $args->{err}->{code}
          $args->{err}->{data}{name}: $args->{err}->{data}{message}
        ERROR
    }

    return;
}

=head2 _normalise_sv_to_int()

Take a scalar value as an argument and make sure it is really an integer value
(See The C<B> module and
L<https://perldoc.perl.org/5.26.1/perlguts#What's-Really-Stored-in-an-SV?|perlguts>
for the gory details).

=cut

sub _normalise_sv_to_int {
    my ($self, $maybe_string_value) = @_;
    return $maybe_string_value + 0;
}


1;
